import argparse
import json
import os
import shutil
import sys
from typing import Optional, List, Sequence, Text

from fyoo.parser import FyooParser


class CliSingleton:

    __instance = None

    DESCRIPTION = '''
    This utility wraps around a command, and templates in context to
    the latter command's arguments. The child process will replace
    the fyoo/python process.
    '''.strip()
    HELP = {
        'exec': '''
        Execute a subcommand. The subcommand will spawn a child process that
        will become a parent (implemented by ``os.execvp``).
        '''.strip(),
        'command': 'Enter any number of arguments as a command.',
        'dry_run': 'Do not actually kick off command.',
        'verbose': 'Show the command before running it.',
    }

    def __init__(self):
        _C = CliSingleton
        self.parser = FyooParser(
            'fyoo', description=_C.DESCRIPTION,
            formatter_class=lambda prog: argparse.ArgumentDefaultsHelpFormatter(prog, width=120))
        self.parser.add_argument('-v', '--verbose', action='store_true', default=False, help=_C.HELP['verbose'])
        self.parser.add_argument('-dr', '--dry-run', action='store_true', default=False, help=_C.HELP['dry_run'])
        subparsers = self.parser.add_subparsers(parser_class=argparse.ArgumentParser)
        subparsers.required = True
        exec_parser = subparsers.add_parser('--', help=_C.HELP['exec'])
        exec_parser.set_defaults(callback=self.exec)
        exec_parser.add_argument('command', nargs=argparse.REMAINDER, help=_C.HELP['command'])

    # pylint: disable=no-self-use
    def exec(self, dry_run: bool, verbose: bool, command: List[str]):
        if dry_run or verbose:
            print(json.dumps(command))
        if not dry_run:
            if shutil.which(command[0]) is None:
                self.parser.error(f"Executable '{command[0]}' does not exist")
            os.execvp(command[0], command)

    def main(self, args: Sequence[Text]) -> None:
        try:
            arg_dict = vars(self.parser.parse_args(args))
        except TypeError as err:
            if err.args == ('sequence item 0: expected str instance, NoneType found', ):
                self.parser.error('Please provide a subcommand')
            raise
        callback = arg_dict.pop('callback')
        callback(**arg_dict)

    def __new__(cls, *args, **kwargs):
        if not kwargs.pop('_is_instance_call', False):
            raise ValueError("Can not instantiate, use .instance() instead")
        return super(CliSingleton, cls).__new__(cls, *args, **kwargs)

    @classmethod
    def instance(cls):
        if cls.__instance is None:
            cls.__instance = cls.__new__(cls, _is_instance_call=True)
            cls.__instance.__init__()
        return cls.__instance

    @classmethod
    def remove(cls):
        if cls.__instance is not None:
            CliSingleton.__instance = None


def get_parser() -> FyooParser:
    return CliSingleton.instance().parser


def main(args: Optional[Sequence[Text]] = None):
    if args is None:
        args = sys.argv[1:]

    cli = CliSingleton.instance()

    cli.main(args)


if __name__ == '__main__':
    main()



"""
Global Attributes
`````````````````

The following attributes are provided to Fyoo's jinja
environment as globals, whether they are objects
or callable functions.

"""
import datetime
import os
import pytz

from fyoo.exception import FyooTemplateException


__all__ = [
    'raw_datetime',
    'date',
    'getenv',
    'throw',
]

raw_datetime = datetime.datetime


def date(tz='UTC', fmt: str = r'%Y-%m-%d') -> str:
    """Get current time string

    .. code-block:: console

       $ fyoo -- echo 'today is {{ date() }}'
       today is 2020-01-01

       $ fyoo -- echo 'the year is {{ date(fmt="%Y") }}'
       the year is 2020


    Args:
        tz (str, optional): Timezone to use. Defaults to 'UTC'.
        fmt (str, optional): Datetime format. Defaults to r'%Y-%m-%d'.

    Returns:
        str: Formated datetime string
    """
    return datetime.datetime.now(tz=pytz.timezone(tz)).strftime(fmt)

getenv = os.getenv
"""Alias of os.getenv

.. code-block:: console

   $ fyoo -- echo 'I am a {{ getenv("USER") }}'
   I am a coolcat
   $ fyoo -- echo 'example {{ getenv("IDONTEXIST", "default") }}'
   example default

"""

    # ""Alias of os.getenv

    # .. code-block:: console

    # $ fyoo -- echo 'I am a {{ getenv("USER") }}
    # I am a coolcat
    # $ fyoo -- echo 'example {{ getenv("IDONTEXIST", "default") }}'
    # example default

    # ""

def throw(*args):
    # pylint: disable=line-too-long
    """Raise a FyooTemplateException

    You would do this if you wanted to verify arguments
    at 'compile-time', before executing a subcommand.

    .. code-block:: console

       $ fyoo -- echo '{% if not table %}{{ throw("table not set") }}{% endif %}table is {{ table }}'
       fyoo.exception.FyooTemplateException: table not set

       $ fyoo --set table=customer -- echo '{% if not table %}{{ throw("table not set") }}{% endif %}table is {{ table }}'
       table is customer

    """
    raise FyooTemplateException(*args)
